/*
 *  Operator.c
 *  MacFM
 *
 *  Created by KIMURA Hiroaki on 10/05/04.
 *  Copyright 2010 __MyCompanyName__. All rights reserved.
 *
 */

#include "Operator.h"

FMInfo* allocFMInfo(){
	
	FMInfo* fmInfo = (FMInfo*)malloc(sizeof(FMInfo));
	fmInfo->carrierFreq = 0;
	fmInfo->freqIndex = 1;
	fmInfo->modIndex = 1;
	fmInfo->isEnabled = 1;
	fmInfo->parents = NULL;
	fmInfo->numOfParents = 0;
	fmInfo->isFB = 0;
	
	//envelope
	fmInfo->attack = 0;
	fmInfo->decay = 0;
	fmInfo->sustain = 1;
	fmInfo->release = 0;
	
	return fmInfo;
}

Speaker* allocSpeaker(){
	Speaker* speaker = (Speaker*)malloc(sizeof(Speaker));
	speaker->numOfParents = 0;
	speaker->fms = NULL;
	
	return speaker;
}

void setCarrierFreq(FMInfo* carrier, double freq){
		
	carrier->carrierFreq = freq;
	
	for (int i = 0; i < carrier->numOfParents; i++){
		FMInfo* parent = carrier->parents[i];
		setCarrierFreq(parent, freq);
	}
	
}

double fmGetPower(FMInfo* fm, double sec){
	
	double power = 1.0;
	
	if (sec < fm->attack){
		power = sec / fm->attack;
	}else if (sec < fm->decay + fm->attack) {
		double t = sec - fm->attack;
		double w = (1.0 - fm->sustain) * t / fm->decay;
		
		power = 1.0 - w;
	}else if( sec < 4){
		power = fm->sustain;
	}else if (sec > 4 && sec < (4 + fm->release)) {
		double t = sec - 4.0;
		power = fm->sustain * (1.0 - t / fm->release);
		
	}else if (sec >  + fm->release){
		power = 0;
	}

	return power;
}



double fmGetValueDefault(FMInfo* fm, double sec){
	double parentPhase = 0.0;
	
	for (int i = 0; i < fm->numOfParents; i++){
		parentPhase += fmGetValueDefault(fm->parents[i], sec);
	}
	
	if (fm->isFB > 0){
		double value = 0.0;
		double beta = M_PI * fm->isFB / 8.0;
		
		for (int i = 0; i < 20; i++){
			value = sin(2.0 * M_PI * (fm->carrierFreq * fm->freqIndex) * sec + beta * value);
		}
		
		return fm->modIndex * value;
		
	}
	
	
	return fm->modIndex * sin(2 * M_PI * (fm->carrierFreq * fm->freqIndex) * sec + parentPhase);
}


double fmGetValue(FMInfo* fm, double sec){
	
	double parentPhase = 0.0;
	
	for (int i = 0; i < fm->numOfParents; i++){
		parentPhase += fmGetValue(fm->parents[i], sec);
	}
	
	if (fm->isFB > 0){
		double value = 0.0;
		double beta = M_PI * fm->isFB / 8.0;
		
		for (int i = 0; i < 20; i++){
			value = sin(2.0 * M_PI * (fm->carrierFreq * fm->freqIndex) * sec + beta * value);
		}
		if (!fm->isEnabled)
			return 0.0;
		return fm->modIndex * value * fmGetPower(fm, sec);
		
	}
	
	
	//disabledの場合は0を返すのではなく、素通しさせる
	if (fm->isEnabled)
		return fm->modIndex * sin(2 * M_PI * (fm->carrierFreq * fm->freqIndex) * sec + parentPhase) * fmGetPower(fm, sec);
	else 
		return fm->modIndex * sin(parentPhase) * fmGetPower(fm, sec);

}

double spGetValue(Speaker* speaker, double sec){
	double value = 0.0;

	for (int i = 0; i < speaker->numOfParents; i++){
		int t = (int)(sec / 5.0);
		double time = sec - 5.0 * t;
		speaker->sec = time;
		value += fmGetValue(speaker->fms[i], time);
	}
	
	if (speaker->numOfParents)
		return value / speaker->numOfParents;
	else 
		return 0.0;

}

void spAddFMInfo(Speaker* speaker, FMInfo* addedFM){
	
	FMInfo** tmpParents = (FMInfo**)malloc(sizeof(FMInfo*) * (speaker->numOfParents + 1));
	
	for (int i = 0; i < speaker->numOfParents; i++){
		tmpParents[i] = speaker->fms[i];
	}
	
	tmpParents[speaker->numOfParents] = addedFM;
	speaker->numOfParents = speaker->numOfParents+1;
	
	if (speaker->fms != NULL)
		free(speaker->fms);
	
	speaker->fms = tmpParents;
	

}

void fmAddFmInfo(FMInfo* fm, FMInfo* addedFM){
	FMInfo** tmpParents = (FMInfo**)malloc(sizeof(FMInfo*) * (fm->numOfParents + 1));
	
	for (int i = 0; i < fm->numOfParents; i++){
		tmpParents[i] = fm->parents[i];
	}
	
	tmpParents[fm->numOfParents] = addedFM;
	fm->numOfParents++;
	
	if (fm->parents != NULL)
		free(fm->parents);
	
	fm->parents = tmpParents;
}


void fmRemoveAllFMInfo(FMInfo* fm){
	if (fm->numOfParents){
		free(fm->parents);
		fm->numOfParents = 0;
		fm->parents = NULL;
	}
	
}

void spRemoveAllFMInfo(Speaker* sp){
	if (sp->numOfParents){
		free(sp->fms);
		sp->numOfParents = 0;
		sp->fms = NULL;
	}
}


void fmRemoveFmInfo(FMInfo* fm, FMInfo* removed){
	int hasFM = 0;
	
	for (int i = 0; i < fm->numOfParents; i++) {
		if(fm->parents[i] == removed){
			hasFM = 1;
		}
	}
	
	if (hasFM){
		FMInfo** tmpInfo = (FMInfo**)malloc(sizeof(FMInfo*) * (fm->numOfParents - 1));
		int index = 0;
		for (int i = 0; i < fm->numOfParents; i++){
			if (fm->parents[i] != removed){
				tmpInfo[index] = fm->parents[i];
				index++;
			}
		}
		free(fm->parents);
		fm->parents = tmpInfo;
		fm->numOfParents--;
	}
}

void spRemoveFmInfo(Speaker* sp, FMInfo* removed){
	int hasFM = 0;
	
	for (int i = 0; i < sp->numOfParents; i++) {
		if(sp->fms[i] == removed){
			hasFM = 1;
		}
	}
	
	if (hasFM){
		FMInfo** tmpInfo = (FMInfo**)malloc(sizeof(FMInfo*) * (sp->numOfParents - 1));
		int index = 0;
		for (int i = 0; i < sp->numOfParents; i++){
			if (sp->fms[i] != removed){
				tmpInfo[index] = sp->fms[i];
				index++;
			}
		}
		free(sp->fms);
		sp->fms = tmpInfo;
		sp->numOfParents--;
	}
}

